/* 
 * linux/include/linux/bossa_policy_rts.h 
 *   Copyright (C) 2004 EMN, DIKU. All rights reserved.
 *   Ecole des Mines de Nantes, University of Copenhagen.
 *
 * This file describes the interface between policies and the RTS.
 *
 */

// The following types describe information defined in the policies and
// used by the RTS

// This is to be included in both the RTS and the policies

#include "bossa_events.h"
#include "bossa_event_headers.h"

struct scheduler_struct;
struct bossa_struct;
struct event_struct;

// only for policies actually...
#ifdef PROTOCOL_BOSSA_STRUCT
#define attr(x) list_entry(x, struct task_struct, bossa)
#endif

#ifndef PROTOCOL_PROCESS_STRUCT
#define PROTOCOL_PROCESS_STRUCT struct bossa_struct
#endif
#ifndef PROTOCOL_SCHEDULER_STRUCT
#define PROTOCOL_SCHEDULER_STRUCT struct scheduler_struct
#endif

/*******************************************************
 * next_list : 
 *
 *     This is the list of virtual/process schedulers managing the
 *     associated process
 *
 *******************************************************/

struct next_list_struct {
  /* the current virtual/process scheduler */
  PROTOCOL_SCHEDULER_STRUCT *sched;
  /* next process in the next list */
  struct next_list_struct *next;
  /* previous process in the next list */
  struct next_list_struct *prev;
};

/*******************************************************
 * scheduler_struct : 
 *
 *     Scheduler operations exported to the RTS
 *     (actually, preempt is only relevant to other policies)
 *
 *******************************************************/

enum sched_type { VS, PS };

struct scheduler_operations
{
  int (*handle_event) (struct event_struct *, // scheduler entry point
		       struct next_list_struct *,
		       struct next_list_struct *);
  int (*block) (int, struct bossa_struct *, struct next_list_struct *);
  int (*block_bossa_schedule) (void);
  int (*block_entry) (int, struct bossa_struct *);
  int (*yield_system_immediate) (int, struct bossa_struct *, struct next_list_struct *);
  int (*yield_system_immediate_bossa_schedule) (void);
  int (*yield_system_immediate_entry) (int, struct bossa_struct *);
  int (*yield_system_pause) (int, struct bossa_struct *, struct next_list_struct *);
  int (*yield_system_pause_bossa_schedule) (void);
  int (*yield_system_pause_entry) (int, struct bossa_struct *);
  int (*unblock_preemptive) (int, struct bossa_struct *, struct next_list_struct *);
  int (*unblock_preemptive_entry) (int, struct bossa_struct *);
  int (*bossa_schedule) (void);
  int (*bossa_schedule_entry) (void);
  int (*unblock_timer_target) (int, struct bossa_struct *, struct next_list_struct *);
  int (*unblock_timer_target_entry) (int, struct bossa_struct *, struct next_list_struct *);
  int (*unblock_timer_notarget) (int, struct bossa_struct *, struct next_list_struct *);
  int (*unblock_timer_notarget_entry) (int, struct bossa_struct *, struct next_list_struct *);
  int (*unblock_timer_persistent)
    (int, struct timer_list * /* the timer itself */,
     void * /* timer policy data */,
     int /* tgt alive */,
     struct bossa_struct *, struct next_list_struct *);
  int (*unblock_timer_persistent_entry)
    (int, struct timer_list * /* the timer itself */,
     void * /* timer policy data */,
     int /* tgt alive */,
     struct bossa_struct *, struct next_list_struct *);
  char *name;
  enum sched_type type;
  int is_default; // true if a default VS/PS
  void (*init) (void);  // scheduler initialization function 
  int (*ioctl_init)(void);
  struct scheduler_struct *
     (*create_scheduler) (struct scheduler_operations *, int new_root);
  unsigned int (*check_process_state) (PROTOCOL_PROCESS_STRUCT *,
				       struct next_list_struct *,
				       unsigned int);
  void (*preempt) (void);
  int (*do_taskprocinfo) (struct next_list_struct *next_list,
			  struct bossa_struct *bossa, char *buffer, int, int,
			  int);
  int (*do_schedprocinfo) (char *buffer, int);
  int (*list_sched_info)(char * page,char ** start,off_t off,
			 int count,int * eof,void * data);
  int (*count_processes)(void);
  void (*print_processes)(void);
  void (*set_priority) (PROTOCOL_PROCESS_STRUCT *,
			struct next_list_struct *,
			int);
  int (*compute_state)(struct next_list_struct *next_list);
  int (*ps_attach)(struct next_list_struct *next_list);
  int (*ps_detach)(struct next_list_struct *next_list, struct bossa_struct *p);
  int (*attach)(PROTOCOL_SCHEDULER_STRUCT *s, int data[]);
  int (*attach_info)(char *info[]);
  int (*detach)(PROTOCOL_SCHEDULER_STRUCT *s, int discard_child);
  void (*attach_self) (void);  // scheduler load-time initialization function 
  PROTOCOL_SCHEDULER_STRUCT *(*detach_self)(int is_root);
  void (*increment_ref_count)(void);
  void (*decrement_ref_count)(void);
  struct next_list_struct next_list;
};

/*******************************************************
 * scheduler_struct : 
 *
 *     used to define Bossa schedulers
 *     every virtual scheduler must define a structure for child schedulers
 *     that begins with this format
 *
 *******************************************************/

struct scheduler_struct {
  /* default attributes for Bossa */
  /* next and previous scheduler in the current bossa queue */
  struct list_head list_info;
  /* scheduler operations exported to the RTS */
  struct scheduler_operations *ops;
  /* bossa schedule state */
  int state;
};

/*******************************************************
 * 
 * event_struct
 *
 *   structure for event notifications
 *
 *******************************************************/

struct event_struct {
  int type;                           /* event type */
  PROTOCOL_PROCESS_STRUCT * target;   /* target process */
  PROTOCOL_PROCESS_STRUCT * source;   /* event process source */
};

/*******************************************************
 * bossa_struct : 
 *
 *     used to define Bossa processes
 *     every process scheduler must define a structure for child processes
 *     that begins with this format
 *
 *    *attr points to the default process atributes (pid, memory context)
 *
 *******************************************************/

#ifndef BOSSA_STRUCT
#define BOSSA_STRUCT
struct bossa_struct {
  /* default attributes for Bossa */
  /* next and previous processes in the current bossa queue */
  struct list_head list_info;
  /* list of virtual schedulers on the way to this process */
  struct next_list_struct *next_list;
  /* process state */
  int state;
};
#endif

/*******************************************************
 *
 * The current state of the scheduling hierarchy
 *
 *******************************************************/

#ifndef SCHED_STATE
#define SCHED_STATE
enum SchedState {
  RUNNING    = 1<<2,
  READY      = 1<<3,
  BLOCKED    = 1<<4,
  TERMINATED = 1<<5
};
extern enum SchedState root_s_state;           // state of the root scheduler
#endif

extern struct bossa_struct *rts_idle_process;  // System wide idle
extern struct bossa_struct *running;

/***************************************************************************/

// The following functions are defined by the RTS and used by policies

// timer functions
struct bossa_timer_data {
  struct bossa_struct *p; // target process, or NULL if global to the policy
  struct next_list_struct *next_list; // next list leading to the policy
  unsigned int timer_index; // indicates which of the policy's timers this is
};

struct bossa_persistent_timer_data {
  struct bossa_struct *p; // target process, or NULL if global to the policy
  struct next_list_struct *next_list; // next list leading to the policy
  unsigned int timer_index; // indicates which of the policy's timers this is
  struct timer_list *timer;
  int target_alive;
};

#ifdef CONFIG_HIGH_RES_TIMERS
extern struct timespec *now(void);
#else
extern unsigned long now(void);
#endif
extern struct timer_list *
create_bossa_timer(struct bossa_struct *p,
		   struct next_list_struct *next_list,
		   unsigned int index);
extern struct timer_list *
create_static_bossa_timer(struct bossa_struct *p,
			  struct next_list_struct *next_list,
			  struct timer_list *timer,
			  struct bossa_timer_data *timer_data,
			  unsigned int index);
extern struct timer_list *
create_persistent_bossa_timer(struct bossa_struct *p,
			      struct next_list_struct *next_list,
			      unsigned int index,
			      int data_sz);
extern void delete_bossa_timer(struct timer_list *t);
extern void delete_static_bossa_timer(struct timer_list *t);
extern void globalize_bossa_timer(struct timer_list *timer);
#ifdef CONFIG_HIGH_RES_TIMERS
extern void start_absolute_timer(struct timer_list *t, struct timespec *abs);
extern void start_relative_timer(struct timer_list *t, struct timespec *rel);
#else
void start_absolute_timer(struct timer_list *t, unsigned long abs_time);
void start_relative_timer(struct timer_list *t, unsigned long rel_time);
#endif
extern void stop_timer(struct timer_list *t);

// copy from user
extern int get_user_int(void *user_int);

// management of modules
int bossa_add_scheduler(struct scheduler_operations *, int stat);
void bossa_remove_scheduler(struct scheduler_operations *);

// other management of schedulers
struct scheduler_operations *find_scheduler_by_id(int position);

// lock for interface functions
extern spinlock_t bossa_scheduler_lock;
