// To be included in sched.c

#ifndef CONFIG_BOSSA_CENTRALIZED
#include <linux/bossa_policy_rts.h>
#endif

// unblocking

#ifndef CONFIG_BOSSA_CENTRALIZED
static
#endif
inline int
rts_unblock(int caller, struct task_struct * p, int synchronous) {
  unsigned long flags;

  /* In Bossa, we only notify an unblock event if the process is not
   * running or in the ready queue In Linux we may have useless/multiple
   * calls to wake_up which may cause the Bossa run-time system to notify
   * an unblock event more than once (an unneeded/inexisting unblock event)
   * */

  p->state = TASK_RUNNING;
  if (synchronous) {
    int res = bossa_check_process_state(&(p->bossa),BLOCKED);
    bossa_notify_event(EVENT_UNBLOCK_NONPREEMPTIVE|caller,(&(p->bossa)),NULL);
    return res;
  }
  else {
    int res;
    spin_lock_irqsave(&bossa_scheduler_lock, flags);
#ifdef CONFIG_BOSSA_LIBRARY
    pre_check_unblock_preemptive(&p->bossa);
#endif
    res = root_scheduler.ops->unblock_preemptive_entry(EVENT_UNBLOCK_PREEMPTIVE|caller,(&(p->bossa)));
#ifdef CONFIG_BOSSA_LIBRARY
    post_check();
    root_s_state = compute_scheduler_state();
#endif
    spin_unlock_irqrestore(&bossa_scheduler_lock, flags);
    // the following should be in the policy...
    if (res && root_s_state == READY) {
      current->need_resched = 1;
    }
    return res;
  }
}


// scheduling functions

#define attr(x) list_entry(x, struct task_struct, bossa)

static inline void yield_system_immediate(int caller) {
#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_yield_schedule(running);
#endif
  root_s_state =
    root_scheduler.ops->yield_system_immediate_entry(caller,running);
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif
}

static inline void yield_system_pause(int caller) {
#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_yield_schedule(running);
#endif
  root_s_state =
    root_scheduler.ops->yield_system_pause_entry(caller,running);
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif
}

static inline void block(int caller) {
#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_block(running);
#endif
  root_s_state = root_scheduler.ops->block_entry(caller,running);
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif
}

static inline void call_bossa_schedule(void) {
#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_call_bossa_schedule();
#endif
  root_s_state = root_scheduler.ops->bossa_schedule_entry();
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif
}

#ifndef CONFIG_BOSSA_CENTRALIZED
static
#endif
inline asmlinkage
struct task_struct *rts_schedule(int caller, struct task_struct *prev)
{
	if (running == rts_idle_process) {
	  call_bossa_schedule();
	}
	else {
	  switch (prev->state) {
	  case TASK_INTERRUPTIBLE:
	    if (!signal_pending(prev)) {
	      block(caller);
	      prev->policy &= ~SCHED_YIELD;
	      break;
	    } // fall through
	    prev->state = TASK_RUNNING;
	  case TASK_RUNNING:
	    if (prev->policy & SCHED_YIELD) {
	      prev->policy &= ~SCHED_YIELD;
	      yield_system_pause(caller); // root state can't be blocked
	      break;
	    }
	    else {
	      switch (root_s_state) {
	      case RUNNING:
		yield_system_immediate(caller);
		break;
	      case READY:
		call_bossa_schedule();
		break;
	      case BLOCKED:
	      default:
		running = rts_idle_process;
		break;
	      }
	      break;
	    }
	  case TASK_UNINTERRUPTIBLE:
	  case TASK_STOPPED:
	  case TASK_ZOMBIE:
	    block(caller);
	    break;
	  }
	}
	return attr(running);
}
