// to be included in kernel/sched.c

#ifndef CONFIG_BOSSA_CENTRALIZED

asmlinkage void schedule_context_switch(void)
{
  struct schedule_data * sched_data;
  struct task_struct *prev, *next;
  int this_cpu;
  struct bossa_struct *local_old_running;

  if (!current->active_mm) BUG();
  need_resched_back:
  prev = current;
  this_cpu = prev->processor;
  
  if (unlikely(in_interrupt())) {
    printk("Scheduling in interrupt\n");
    BUG();
  }
  
  release_kernel_lock(prev, this_cpu);

#ifdef CONFIG_BOSSA_TRACE
  get_jinfo();
#endif
  
  /*
   * 'sched_data' is protected by the fact that we can run
   * only one process per CPU.
   */
  sched_data = & aligned_data[this_cpu].schedule_data;
  
  // We are here only if there is a new process to run
  
  // Warning switch_to is not a proc.. prev and next are mandatory          
  next = attr(running);
  
  /*
   * from this point on nothing can prevent us from
   * switching to the next task, save this fact in
   * sched_data.
   */
  sched_data->curr = next;
  kstat.context_swtch++;
  /*
   * there are 3 processes which are affected by a context switch:
   *
   * prev == .... ==> (last => next)
   *
   * It's the 'much more previous' 'prev' that is on next's stack,
   * but prev is set to (the just run) 'last' process by switch_to().
   * This might sound slightly confusing but makes tons of sense.
   */
  
  prepare_to_switch();
  {
    struct mm_struct *mm = next->mm;
    struct mm_struct *oldmm = prev->active_mm;
    if (!mm) {
      if (next->active_mm) BUG();
      next->active_mm = oldmm;
      atomic_inc(&oldmm->mm_count);
      enter_lazy_tlb(oldmm, next, this_cpu);
    } else {
      if (next->active_mm != mm) BUG();
      switch_mm(oldmm, mm, next, this_cpu);
    }
    
    if (!prev->mm) {
      prev->active_mm = NULL;
      mmdrop(oldmm);
    }
  }
  
  /*
   * This just switches the register state and the
   * stack.
   */
  
  switch_to(prev, next, prev);
  __schedule_tail(prev);
  
  outbossa:
  reacquire_kernel_lock(current);
  // we check the bossa state, because that could have changed after
  // the new process is chosen, even though current is not yet
  // updated to the new process at that point
  // anyway, the state seems to be always READY when need_resched
  // is set, except in init/main.c, which is not relevant here
  if (root_s_state == READY /*current->need_resched*/) {
    spin_lock_irq(&bossa_scheduler_lock);
    local_old_running = running;
    bossa_schedule();
    attr(local_old_running)->need_resched = 0;
    spin_unlock_irq(&bossa_scheduler_lock);
    if (local_old_running != running)
      goto need_resched_back;
  }
  return;
}

asmlinkage void schedule_from_idle(int caller) {
  struct bossa_struct *local_old_running;
  
  /* Bossa RTS begins here */
  spin_lock_irq(&bossa_scheduler_lock);
  local_old_running = running;
  
// end prelude, choose process

  bossa_schedule();

// postlude
  current->need_resched = 0;
  spin_unlock_irq(&bossa_scheduler_lock);
  if (running != local_old_running) schedule_context_switch();
}

asmlinkage void schedule_uninterruptible(int caller) {
  struct bossa_struct *local_old_running;
  
  /* Bossa RTS begins here */
  spin_lock_irq(&bossa_scheduler_lock);
  local_old_running = running;
  
// end prelude, choose process
  
  if (current->state == TASK_RUNNING) {
    switch (root_s_state) {
    case RUNNING:
      yield_system_immediate(caller);
      break;
    case READY:
      bossa_schedule();
      break;
    case BLOCKED:
    default:
      running = rts_idle_process;
      break;
    }
  }
  else block(caller);

// postlude
  current->need_resched = 0;
  spin_unlock_irq(&bossa_scheduler_lock);
  if (running != local_old_running) schedule_context_switch();
}

asmlinkage void schedule_interruptible(int caller) {
  struct bossa_struct *local_old_running;
  
  /* Bossa RTS begins here */
  spin_lock_irq(&bossa_scheduler_lock);
  local_old_running = running;
  
// end prelude, choose process
  
  if (current->state == TASK_INTERRUPTIBLE) {
    if (!signal_pending(current)) {
      block(caller);
      goto schedule_interruptible_out;
    }
    else current->state = TASK_RUNNING;
  }
  switch (root_s_state) {
  case RUNNING:
    yield_system_immediate(caller);
    break;
  case READY:
    bossa_schedule();
    break;
  case BLOCKED:
  default:
    running = rts_idle_process;
    break;
  }
// postlude
schedule_interruptible_out:
  current->need_resched = 0;
  spin_unlock_irq(&bossa_scheduler_lock);
  if (running != local_old_running) schedule_context_switch();
}

// not optimized, but doesn't occur much either
asmlinkage void schedule_uninterruptible_interruptible(int caller) {
  if (current->state == TASK_INTERRUPTIBLE)
    schedule_interruptible(caller);
  else schedule_uninterruptible(caller);
}    

asmlinkage void schedule_yield_immediate(int caller) {
  struct bossa_struct *local_old_running;

  /* Bossa RTS begins here */
  spin_lock_irq(&bossa_scheduler_lock);
  local_old_running = running;
  
// end prelude, choose process

  switch (root_s_state) {
  case RUNNING:
    yield_system_immediate(caller);
    break;
    // fall through
  case READY:
    bossa_schedule();
    break;
  case BLOCKED:
  default:
    running = rts_idle_process;
    break;
  }
  
// postlude
  current->need_resched = 0;
  spin_unlock_irq(&bossa_scheduler_lock);
  if (running != local_old_running) schedule_context_switch();
}

asmlinkage void schedule_from_entry(void) {
  schedule_yield_immediate(0);
}

asmlinkage void schedule_interruptible_pause(int caller) {
  struct bossa_struct *local_old_running;
  
  /* Bossa RTS begins here */
  spin_lock_irq(&bossa_scheduler_lock);
  local_old_running = running;
  
// end prelude, choose process
  
  if (current->state == TASK_INTERRUPTIBLE) {
    if (!signal_pending(current)) {
      block(caller);
      goto schedule_interruptible_out;
    }
    else current->state = TASK_RUNNING;
  }
  if (root_s_state == BLOCKED)
    running = rts_idle_process;
  else yield_system_pause(caller);

// postlude
schedule_interruptible_out:
  current->need_resched = 0;
  spin_unlock_irq(&bossa_scheduler_lock);
  if (running != local_old_running) schedule_context_switch();
}

asmlinkage void schedule_yield_pause(int caller) {
  struct bossa_struct *local_old_running;

  /* Bossa RTS begins here */
  spin_lock_irq(&bossa_scheduler_lock);
  local_old_running = running;
  
// end prelude, choose process
  if (root_s_state == BLOCKED)
    running = rts_idle_process;
  else yield_system_pause(caller);
  
// postlude
  current->need_resched = 0;
  spin_unlock_irq(&bossa_scheduler_lock);
  if (running != local_old_running) schedule_context_switch();
}

// schedule timeouts -------------------------------------------------------

static void process_timeout(unsigned long __data);

signed long schedule_timeout_yield_immediate(int caller, signed long timeout)
{
	struct timer_list timer;
	unsigned long expire;
	//int z,zz;
	switch (timeout)
	{
	case MAX_SCHEDULE_TIMEOUT:
		/*
		 * These two special cases are useful to be comfortable
		 * in the caller. Nothing more. We could take
		 * MAX_SCHEDULE_TIMEOUT from one of the negative value
		 * but I' d like to return a valid offset (>=0) to allow
		 * the caller to do everything it want with the retval.
		 */
		{
		B_PRINTK("schedule_timeout: calling schedule().No event here\n");
		schedule_yield_immediate(caller); // timeout too high, set no timer.
		goto out;
		}
	default:
		/*
		 * Another bit of PARANOID. Note that the retval will be
		 * 0 since no piece of kernel is supposed to do a check
		 * for a negative retval of schedule_timeout() (since it
		 * should never happens anyway). You just have the printk()
		 * that will tell you if something is gone wrong and where.
		 */
		if (timeout < 0)
		{
			printk(KERN_ERR "schedule_timeout: wrong timeout "
			       "value %lx from %p\n", timeout,
			       __builtin_return_address(0));
			current->state = TASK_RUNNING;
			BOSSA_UNBLOCK_RUNNING(TIMER, current);
			goto out;
		}
	}

	expire = timeout + jiffies;

	init_timer(&timer);
	timer.expires = expire;
	timer.data = (unsigned long) current;
	timer.function = process_timeout;

	add_timer(&timer);
	 // change event name -> block_wait_queue _no_signal ?
        schedule_yield_immediate(caller);
	del_timer_sync(&timer);

	timeout = expire - jiffies;

 out:
	return timeout < 0 ? 0 : timeout;
}

signed long schedule_timeout_interruptible(int caller, signed long timeout)
{
	struct timer_list timer;
	unsigned long expire;
	//int z,zz;
	switch (timeout)
	{
	case MAX_SCHEDULE_TIMEOUT:
		/*
		 * These two special cases are useful to be comfortable
		 * in the caller. Nothing more. We could take
		 * MAX_SCHEDULE_TIMEOUT from one of the negative value
		 * but I' d like to return a valid offset (>=0) to allow
		 * the caller to do everything it want with the retval.
		 */
		{
		B_PRINTK("schedule_timeout: calling schedule().No event here\n");
		schedule_interruptible(caller); // timeout too high, set no timer.
		goto out;
		}
	default:
		/*
		 * Another bit of PARANOID. Note that the retval will be
		 * 0 since no piece of kernel is supposed to do a check
		 * for a negative retval of schedule_timeout() (since it
		 * should never happens anyway). You just have the printk()
		 * that will tell you if something is gone wrong and where.
		 */
		if (timeout < 0)
		{
			printk(KERN_ERR "schedule_timeout: wrong timeout "
			       "value %lx from %p\n", timeout,
			       __builtin_return_address(0));
			current->state = TASK_RUNNING;
			BOSSA_UNBLOCK_RUNNING(TIMER, current);
			goto out;
		}
	}

	expire = timeout + jiffies;

	init_timer(&timer);
	timer.expires = expire;
	timer.data = (unsigned long) current;
	timer.function = process_timeout;

	add_timer(&timer);
	 // change event name -> block_wait_queue _no_signal ?
        schedule_interruptible(caller);
	del_timer_sync(&timer);

	timeout = expire - jiffies;

 out:
	return timeout < 0 ? 0 : timeout;
}

signed long schedule_timeout_uninterruptible(int caller, signed long timeout)
{
	struct timer_list timer;
	unsigned long expire;
	//int z,zz;
	switch (timeout)
	{
	case MAX_SCHEDULE_TIMEOUT:
		/*
		 * These two special cases are useful to be comfortable
		 * in the caller. Nothing more. We could take
		 * MAX_SCHEDULE_TIMEOUT from one of the negative value
		 * but I' d like to return a valid offset (>=0) to allow
		 * the caller to do everything it want with the retval.
		 */
		{
		schedule_uninterruptible(caller); // timeout too high, set no timer.
		goto out;
		}
	default:
		/*
		 * Another bit of PARANOID. Note that the retval will be
		 * 0 since no piece of kernel is supposed to do a check
		 * for a negative retval of schedule_timeout() (since it
		 * should never happens anyway). You just have the printk()
		 * that will tell you if something is gone wrong and where.
		 */
		if (timeout < 0)
		{
			printk(KERN_ERR "schedule_timeout: wrong timeout "
			       "value %lx from %p\n", timeout,
			       __builtin_return_address(0));
			current->state = TASK_RUNNING;
			BOSSA_UNBLOCK_RUNNING(TIMER, current);
			goto out;
		}
	}

	expire = timeout + jiffies;

	init_timer(&timer);
	timer.expires = expire;
	timer.data = (unsigned long) current;
	timer.function = process_timeout;

	add_timer(&timer);
	 // change event name -> block_wait_queue _no_signal ?
        schedule_uninterruptible(caller);
	del_timer_sync(&timer);

	timeout = expire - jiffies;

 out:
	return timeout < 0 ? 0 : timeout;
}

/*
EXPORT_SYMBOL(schedule_from_entry);
EXPORT_SYMBOL(schedule_from_idle);
EXPORT_SYMBOL(schedule_uninterruptible);
EXPORT_SYMBOL(schedule_uninterruptible_interruptible);
EXPORT_SYMBOL(schedule_interruptible);
EXPORT_SYMBOL(schedule_interruptible_pause);
EXPORT_SYMBOL(schedule_yield_immediate);
EXPORT_SYMBOL(schedule_yield_pause);
EXPORT_SYMBOL(schedule_timeout_yield_immediate);
EXPORT_SYMBOL(schedule_timeout_uninterruptible);
EXPORT_SYMBOL(schedule_timeout_interruptible);
*/
#endif
